from typing import (TYPE_CHECKING, Any, Awaitable, Callable, List, Mapping, MutableMapping, Tuple,
                    TypeVar, Union)

if TYPE_CHECKING:
    from .request import Request

TASGIMessage = Mapping[str, Any]
TASGISend = Callable[[TASGIMessage], Awaitable[None]]
TASGIReceive = Callable[[], Awaitable[TASGIMessage]]
TASGIScope = MutableMapping[str, Any]
TASGIHeaders = List[Tuple[bytes, bytes]]
TASGIApp = Callable[[TASGIScope, TASGIReceive, TASGISend], Awaitable[Any]]

TResponseContent = Union[bytes, str]
TMiddleware = Callable[[TASGIScope, TASGIReceive, TASGISend], Awaitable[Any]]
TJSON = Union[None, bool, int, float, str, List["TJSON"], Mapping[str, "TJSON"]]
TExceptionHandler = Callable[["Request", BaseException], Awaitable]

TV = TypeVar("TV")
TVCallable = TypeVar("TVCallable", bound=Callable)
TVAsyncCallable = TypeVar("TVAsyncCallable", bound=Callable[..., Awaitable])
TVExceptionHandler = TypeVar("TVExceptionHandler", bound=TExceptionHandler)



"""ASGI-Tools Utils."""

import re
from functools import wraps
from inspect import isasyncgenfunction, iscoroutinefunction
from typing import Awaitable, Callable, Dict, Tuple, overload
from urllib.parse import unquote_to_bytes

from multidict import CIMultiDict

from asgi_tools import BASE_ENCODING
from asgi_tools.types import TV, TASGIHeaders, TVAsyncCallable


def is_awaitable(fn: Callable) -> bool:
    """Check than the given function is awaitable."""
    return iscoroutinefunction(fn) or isasyncgenfunction(fn)


@overload
def to_awaitable(fn: TVAsyncCallable) -> TVAsyncCallable:
    ...


@overload
def to_awaitable(fn: Callable[..., TV]) -> Callable[..., Awaitable[TV]]:
    ...


def to_awaitable(fn: Callable):
    """Convert the given function to a coroutine function if it isn't"""
    if is_awaitable(fn):
        return fn

    @wraps(fn)
    async def coro(*args, **kwargs):
        return fn(*args, **kwargs)

    return coro


def parse_headers(headers: TASGIHeaders) -> CIMultiDict:
    """Decode the given headers list."""
    return CIMultiDict(
        [(n.decode(BASE_ENCODING), v.decode(BASE_ENCODING)) for n, v in headers]
    )


OPTION_HEADER_PIECE_RE = re.compile(
    r"""
    \s*,?\s*  # newlines were replaced with commas
    (?P<key>
        "[^"\\]*(?:\\.[^"\\]*)*"  # quoted string
    |
        [^\s;,=*]+  # token
    )
    (?:\*(?P<count>\d+))?  # *1, optional continuation index
    \s*
    (?:  # optionally followed by =value
        (?:  # equals sign, possibly with encoding
            \*\s*=\s*  # * indicates extended notation
            (?:  # optional encoding
                (?P<encoding>[^\s]+?)
                '(?P<language>[^\s]*?)'
            )?
        |
            =\s*  # basic notation
        )
        (?P<value>
            "[^"\\]*(?:\\.[^"\\]*)*"  # quoted string
        |
            [^;,]+  # token
        )?
    )?
    \s*;?
    """,
    flags=re.VERBOSE,
)


def parse_options_header(value: str) -> Tuple[str, Dict[str, str]]:
    """Parse the given content disposition header."""

    options: Dict[str, str] = {}
    if not value:
        return "", options

    if ";" not in value:
        return value, options

    ctype, rest = value.split(";", 1)
    while rest:
        match = OPTION_HEADER_PIECE_RE.match(rest)
        if not match:
            break

        option, count, encoding, _, value = match.groups()
        if value is not None:

            if encoding is not None:
                value = unquote_to_bytes(value).decode(encoding)

            if count:
                value = options.get(option, "") + value

        options[option] = value.strip('" ').replace("\\\\", "\\").replace('\\"', '"')
        rest = rest[match.end() :]  # noqa

    return ctype, options



"""Simple Base for ASGI Apps."""

import logging
from functools import partial
from inspect import isclass
from typing import Callable, Dict, List, Optional, Set, Type, Union

from http_router import PrefixedRoute
from http_router.types import TMethods, TPath

from asgi_tools.router import Router

from . import ASGIConnectionClosed, ASGIMethodNotAllowed, ASGINotFound, asgi_logger
from .middleware import LifespanMiddleware, StaticFilesMiddleware, parse_response
from .request import Request
from .response import Response, ResponseError
from .types import (TASGIReceive, TASGIScope, TASGISend, TExceptionHandler, TVCallable,
                    TVExceptionHandler)
from .utils import iscoroutinefunction, to_awaitable


class App:
    """A helper to build ASGI Applications.

    Features:

    * Routing
    * ASGI-Tools :class:`Request`, :class:`Response`
    * Exception management
    * Static files
    * Lifespan events
    * Simplest middlewares

    :param debug: Enable debug mode (more logging, raise unhandled exceptions)
    :type debug: bool, False

    :param logger: Custom logger for the application
    :type logger: logging.Logger

    :param static_url_prefix: A prefix for static files
    :type static_url_prefix: str, "/static"

    :param static_folders: A list of folders to look static files
    :type static_folders: list[str]

    :param trim_last_slash: Consider "/path" and "/path/" as the same
    :type trim_last_slash: bool, False

    """

    exception_handlers: Dict[Type[BaseException], TExceptionHandler]

    def __init__(
        self,
        *,
        debug: bool = False,
        logger: logging.Logger = asgi_logger,
        static_url_prefix: str = "/static",
        static_folders: Union[str, List[str], None] = None,
        trim_last_slash: bool = False,
    ):
        """Initialize router and lifespan middleware."""

        # Register base exception handlers
        self.exception_handlers = {
            ASGIConnectionClosed: to_awaitable(lambda _, __: None)
        }

        # Setup routing
        self.router = Router(
            trim_last_slash=trim_last_slash, validator=callable, converter=to_awaitable
        )

        # Setup logging
        self.logger = logger

        # Setup app
        self.__app__ = self.__match__

        # Setup lifespan
        self.lifespan = LifespanMiddleware(
            self.__process__, ignore_errors=not debug, logger=self.logger
        )

        # Enable middleware for static files
        if static_folders and static_url_prefix:
            md = StaticFilesMiddleware.setup(
                folders=static_folders, url_prefix=static_url_prefix
            )
            self.middleware(md)

        # Debug mode
        self.debug = debug

        # Handle unknown exceptions
        if not debug:

            async def handle_unknown_exception(
                _: Request, exc: BaseException
            ) -> Response:
                self.logger.exception(exc)
                return ResponseError.INTERNAL_SERVER_ERROR()

            self.exception_handlers[Exception] = handle_unknown_exception

        self.internal_middlewares: List = []

    async def __call__(
        self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend
    ) -> None:
        """Convert the given scope into a request and process."""
        await self.lifespan(scope, receive, send)

    async def __process__(
        self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend
    ):
        """Send ASGI messages."""
        scope["app"] = self
        request = Request(scope, receive, send)
        try:
            response: Optional[Response] = await self.__app__(request, receive, send)
            if response is not None:
                await response(scope, receive, send)

        # Handle exceptions
        except BaseException as exc:  # noqa
            for etype in type(exc).mro():
                if etype in self.exception_handlers:
                    await parse_response(
                        await self.exception_handlers[etype](request, exc)
                    )(scope, receive, send)
                    break
            else:
                if isinstance(exc, Response):
                    await exc(scope, receive, send)
                else:
                    raise

    async def __match__(
        self, request: Request, _: TASGIReceive, send: TASGISend
    ) -> Optional[Response]:
        """Find and call a callback, parse a response, handle exceptions."""
        scope = request.scope
        path = f"{ scope.get('root_path', '') }{ scope['path'] }"
        try:
            match = self.router(path, scope.get("method", "GET"))

        except ASGINotFound as exc:
            raise ResponseError.NOT_FOUND() from exc

        except ASGIMethodNotAllowed as exc:
            raise ResponseError.METHOD_NOT_ALLOWED() from exc

        scope["path_params"] = {} if match.params is None else match.params
        response = await match.target(request)
        scope_type = scope["type"]

        if scope_type == "http":
            return parse_response(response)

        # TODO: Do we need to close websockets automatically?
        # if scope_type == "websocket":
        #     await send({"type": "websocket.close"})

        return None

    def __route__(self, router: Router, *prefixes: str, **_) -> "App":
        """Mount self as a nested application."""
        for prefix in prefixes:
            route = RouteApp(prefix, set(), target=self)
            router.dynamic.insert(0, route)
        return self

    def middleware(self, md: TVCallable, insert_first: bool = False) -> TVCallable:
        """Register a middleware."""
        # Register as a simple middleware
        if iscoroutinefunction(md):

            if md not in self.internal_middlewares:
                if insert_first:
                    self.internal_middlewares.insert(0, md)
                else:
                    self.internal_middlewares.append(md)

            app = self.__match__
            for md in reversed(self.internal_middlewares):
                app = partial(md, app)

            self.__app__ = app

        else:
            self.lifespan.bind(md(self.lifespan.app))

        return md

    def route(self, *paths: TPath, methods: Optional[TMethods] = None, **opts):
        """Register a route."""
        return self.router.route(*paths, methods=methods, **opts)

    def on_startup(self, fn: Callable) -> None:
        """Register a startup handler."""
        return self.lifespan.on_startup(fn)

    def on_shutdown(self, fn: Callable) -> None:
        """Register a shutdown handler."""
        return self.lifespan.on_shutdown(fn)

    def on_error(self, etype: Type[BaseException]):
        """Register an exception handler.

        .. code-block::

            @app.on_error(TimeoutError)
            async def timeout(request, error):
                return 'Something bad happens'

            @app.on_error(ResponseError)
            async def process_http_errors(request, response_error):
                if response_error.status_code == 404:
                    return render_template('page_not_found.html'), 404
                return response_error

        """
        assert isclass(etype) and issubclass(
            etype, BaseException
        ), f"Invalid exception type: {etype}"

        def recorder(handler: TVExceptionHandler) -> TVExceptionHandler:
            self.exception_handlers[etype] = to_awaitable(handler)
            return handler

        return recorder


class RouteApp(PrefixedRoute):
    """Custom route to submount an application."""

    def __init__(self, path: str, methods: Set, target: App):
        """Create app callable."""
        path = path.rstrip("/")

        def app(request: Request):
            subrequest = request.__copy__(path=request.path[len(path) :])  # noqa
            return target.__app__(subrequest, request.receive, request.send)

        super().__init__(path, methods, app)



"""Testing tools."""

import binascii
import io
import mimetypes
import os
import random
from collections import deque
from contextlib import asynccontextmanager, suppress
from functools import partial
from http.cookies import SimpleCookie
from json import loads
from pathlib import Path
from typing import (Any, AsyncGenerator, Awaitable, Callable, Coroutine, Deque, Dict, Mapping,
                    Optional, Tuple, Union, cast)
from urllib.parse import urlencode

from multidict import MultiDict
from yarl import URL

from asgi_tools import BASE_ENCODING, DEFAULT_CHARSET, ASGIConnectionClosed
from asgi_tools._compat import FIRST_COMPLETED, aio_cancel, aio_sleep, aio_spawn, aio_wait
from asgi_tools.response import Response, ResponseJSON, ResponseWebSocket, parse_websocket_msg
from asgi_tools.types import TJSON, TASGIApp, TASGIMessage, TASGIReceive, TASGIScope, TASGISend
from asgi_tools.utils import CIMultiDict, parse_headers


class TestResponse(Response):
    """Response for test client."""

    def __init__(
        self,
    ):
        super().__init__(b"")

    async def __call__(self, _: TASGIScope, receive: TASGIReceive, __: TASGISend):
        self._receive = receive
        msg = await self._receive()
        assert msg.get("type") == "http.response.start", "Invalid Response"
        self.status_code = int(msg.get("status", 502))
        self.headers = cast(MultiDict, parse_headers(msg.get("headers", [])))
        self.content_type = self.headers.get("content-type")
        for cookie in self.headers.getall("set-cookie", []):
            self.cookies.load(cookie)

    async def stream(self) -> AsyncGenerator[bytes, None]:
        """Stream the response."""
        more_body = True
        while more_body:
            msg = await self._receive()
            if msg.get("type") == "http.response.body":
                chunk = msg.get("body")
                if chunk:
                    yield chunk
                more_body = msg.get("more_body", False)

    async def body(self) -> bytes:
        """Load response body."""
        body_ = b""
        async for chunk in self.stream():
            body_ += chunk
        return body_

    async def text(self) -> str:
        body = await self.body()
        return body.decode(DEFAULT_CHARSET)

    async def json(self) -> TJSON:
        text = await self.text()
        return loads(text)


class TestWebSocketResponse(ResponseWebSocket):
    """Support websockets in tests."""

    def connect(self) -> Coroutine[TASGIMessage, Any, Any]:
        return self.send({"type": "websocket.connect"})

    async def disconnect(self):
        await self.send({"type": "websocket.disconnect", "code": 1005})
        self.state = self.STATES.DISCONNECTED

    def send(self, msg, type="websocket.receive"):  # noqa
        """Send a message to a client."""
        return super().send(msg, type=type)

    async def receive(self, raw=False):
        """Receive messages from a client."""
        if self.partner_state == self.STATES.DISCONNECTED:
            raise ASGIConnectionClosed

        msg = await self._receive()
        if not msg["type"].startswith("websocket."):
            raise ValueError(f"Invalid websocket message: {msg!r}")

        if msg["type"] == "websocket.accept":
            self.partner_state = self.STATES.CONNECTED
            return await self.receive(raw=raw)

        if msg["type"] == "websocket.close":
            self.partner_state = self.STATES.DISCONNECTED
            raise ASGIConnectionClosed("Connection has been closed.")

        return msg if raw else parse_websocket_msg(msg, charset=DEFAULT_CHARSET)


class ASGITestClient:
    """The test client allows you to make requests against an ASGI application.

    Features:

    * cookies
    * multipart/form-data
    * follow redirects
    * request streams
    * response streams
    * websocket support
    * lifespan management

    """

    def __init__(self, app: TASGIApp, base_url: str = "http://localhost"):
        self.app = app
        self.base_url = URL(base_url)
        self.cookies: SimpleCookie = SimpleCookie()
        self.headers: Dict[str, str] = {}

    def __getattr__(self, name: str) -> Callable[..., Awaitable]:
        return partial(self.request, method=name.upper())

    async def request(
        self,
        path: str,
        method: str = "GET",
        query: Union[str, Dict] = "",
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Dict[str, str]] = None,
        data: Union[bytes, str, Dict, AsyncGenerator] = b"",
        json: TJSON = None,
        follow_redirect: bool = True,
        timeout: float = 10.0,
    ) -> TestResponse:
        """Make a HTTP requests."""

        headers = headers or dict(self.headers)

        if isinstance(data, str):
            data = Response.process_content(data)

        elif isinstance(data, dict):
            is_multipart = any(isinstance(value, io.IOBase) for value in data.values())
            if is_multipart:
                data, headers["Content-Type"] = encode_multipart(data)

            else:
                headers["Content-Type"] = "application/x-www-form-urlencoded"
                data = urlencode(data).encode(DEFAULT_CHARSET)

        elif json is not None:
            headers["Content-Type"] = "application/json"
            data = ResponseJSON.process_content(json)

        pipe = Pipe()

        if isinstance(data, bytes):
            headers.setdefault("Content-Length", str(len(data)))

        scope = self.build_scope(
            path,
            headers=headers,
            query=query,
            cookies=cookies,
            type="http",
            method=method,
        )

        await aio_wait(
            aio_wait(
                stream_data(data, pipe.send_to_app),
                self.app(scope, pipe.receive_from_app, pipe.send_to_client),
            ),
            raise_timeout(timeout),
            strategy=FIRST_COMPLETED,
        )

        res = TestResponse()
        await res(scope, pipe.receive_from_client, pipe.send_to_app)
        for n, v in res.cookies.items():
            self.cookies[n] = v

        if follow_redirect and res.status_code in {301, 302, 303, 307, 308}:
            return await self.get(res.headers["location"])

        return res

    # TODO: Timeouts for websockets
    @asynccontextmanager
    async def websocket(
        self,
        path: str,
        query: Union[str, Dict, None] = None,
        headers: Optional[Dict] = None,
        cookies: Optional[Dict] = None,
    ):
        """Connect to a websocket."""
        pipe = Pipe()

        ci_headers = CIMultiDict(headers or {})

        scope = self.build_scope(
            path,
            headers=ci_headers,
            query=query,
            cookies=cookies,
            type="websocket",
            subprotocols=str(ci_headers.get("Sec-WebSocket-Protocol", "")).split(","),
        )
        ws = TestWebSocketResponse(scope, pipe.receive_from_client, pipe.send_to_app)
        async with aio_spawn(
            self.app, scope, pipe.receive_from_app, pipe.send_to_client
        ):
            await ws.connect()
            yield ws
            await ws.disconnect()

    def lifespan(self, timeout: float = 3e-2):
        """Manage `Lifespan <https://asgi.readthedocs.io/en/latest/specs/lifespan.html>`_
        protocol."""
        return manage_lifespan(self.app, timeout=timeout)

    def build_scope(
        self,
        path: str,
        headers: Union[Dict, CIMultiDict, None] = None,
        query: Union[str, Dict, None] = None,
        cookies: Optional[Dict] = None,
        **scope,
    ) -> TASGIScope:
        """Prepare a request scope."""
        headers = headers or {}
        headers.setdefault("User-Agent", "ASGI-Tools-Test-Client")
        headers.setdefault("Host", self.base_url.host)

        if cookies:
            for c, v in cookies.items():
                self.cookies[c] = v

        if len(self.cookies):
            headers.setdefault("Cookie", self.cookies.output(header="", sep=";"))

        url = URL(path)
        if query:
            url = url.with_query(query)

        # Setup client
        scope.setdefault("client", ("127.0.0.1", random.randint(1024, 65535)))

        return dict(
            {
                "asgi": {"version": "3.0"},
                "http_version": "1.1",
                "path": url.path,
                "query_string": url.raw_query_string.encode(),
                "raw_path": url.raw_path.encode(),
                "root_path": "",
                "scheme": scope.get("type") == "http" and self.base_url.scheme or "ws",
                "headers": [
                    (key.lower().encode(BASE_ENCODING), str(val).encode(BASE_ENCODING))
                    for key, val in (headers or {}).items()
                ],
                "server": ("127.0.0.1", self.base_url.port),
            },
            **scope,
        )


def encode_multipart(data: Dict) -> Tuple[bytes, str]:
    body = io.BytesIO()
    boundary = binascii.hexlify(os.urandom(16))
    for name, value in data.items():
        headers = f'Content-Disposition: form-data; name="{ name }"'
        if hasattr(value, "read"):
            filename = getattr(value, "name", None)
            if filename:
                headers = f'{ headers }; filename="{ Path(filename).name }"'
                content_type = (
                    mimetypes.guess_type(filename)[0] or "application/octet-stream"
                )
                headers = f"{ headers }\r\nContent-Type: { content_type }"
            value = value.read()

        body.write(b"--%b\r\n" % boundary)
        body.write(headers.encode("utf-8"))
        body.write(b"\r\n\r\n")
        if isinstance(value, str):
            value = value.encode("utf-8")
        body.write(value)
        body.write(b"\r\n")

    body.write(b"--%b--\r\n" % boundary)
    return body.getvalue(), (b"multipart/form-data; boundary=%s" % boundary).decode()


class Pipe:

    __slots__ = (
        "delay",
        "app_is_closed",
        "client_is_closed",
        "app_queue",
        "client_queue",
    )

    def __init__(self, delay: float = 1e-3):
        self.delay = delay
        self.app_is_closed = False
        self.client_is_closed = False
        self.app_queue: Deque[TASGIMessage] = deque()
        self.client_queue: Deque[TASGIMessage] = deque()

    async def send_to_client(self, msg: TASGIMessage):
        if self.client_is_closed:
            raise RuntimeError(f"Unexpected ASGI message to client '{msg.get('type')}'")

        if msg.get("type") == "websocket.close":
            self.client_is_closed = True

        elif msg.get("type") == "http.response.body":
            self.client_is_closed = not msg.get("more_body", False)

        self.client_queue.append(msg)

    async def send_to_app(self, msg: TASGIMessage):
        if self.app_is_closed:
            raise RuntimeError(f"Unexpected ASGI message to app '{msg.get('type')}'")

        if msg.get("type") == "http.disconnect":
            self.app_is_closed = True

        self.app_queue.append(msg)

    async def receive_from_client(self):
        while not self.client_queue:
            await aio_sleep(self.delay)
        return self.client_queue.popleft()

    async def receive_from_app(self):
        while not self.app_queue:
            await aio_sleep(self.delay)
        return self.app_queue.popleft()


async def raise_timeout(timeout: Union[int, float]):
    await aio_sleep(timeout)
    raise TimeoutError("Timeout occured")


async def stream_data(
    data: Union[bytes, AsyncGenerator[Any, bytes]],
    send: Callable[[TASGIMessage], Awaitable[None]],
):
    """Stream a data to an application."""

    if isinstance(data, bytes):
        return await send({"type": "http.request", "body": data, "more_body": False})

    async for chunk in data:
        await send({"type": "http.request", "body": chunk, "more_body": True})
    await send({"type": "http.request", "body": b"", "more_body": False})


@asynccontextmanager
async def manage_lifespan(app, timeout: float = 3e-2):
    """Manage `Lifespan <https://asgi.readthedocs.io/en/latest/specs/lifespan.html>`_ protocol."""
    pipe = Pipe()

    scope = {"type": "lifespan"}

    async def safe_spawn():
        with suppress(BaseException):
            await app(scope, pipe.receive_from_app, pipe.send_to_client)

    async with aio_spawn(safe_spawn) as task:
        await pipe.send_to_app({"type": "lifespan.startup"})
        msg = await aio_wait(
            pipe.receive_from_client(), aio_sleep(timeout), strategy=FIRST_COMPLETED
        )
        if msg and isinstance(msg, Mapping):
            if msg["type"] == "lifespan.startup.failed":
                await aio_cancel(task)
            else:
                assert msg["type"] == "lifespan.startup.complete"

        yield

        await pipe.send_to_app({"type": "lifespan.shutdown"})
        msg = await aio_wait(
            pipe.receive_from_client(), aio_sleep(timeout), strategy=FIRST_COMPLETED
        )
        if msg and isinstance(msg, Mapping):
            assert msg["type"] == "lifespan.shutdown.complete"


# pylama:ignore=D



""" ASGI-Tools -- Tools to make ASGI Applications """

__version__ = "0.70.4"
__license__ = "MIT"

import logging

asgi_logger: logging.Logger = logging.getLogger("asgi-tools")


class ASGIError(Exception):
    """Base class for ASGI-Tools Errors."""


class ASGIConnectionClosed(ASGIError):
    """ASGI-Tools connection closed error."""


class ASGIDecodeError(ASGIError, ValueError):
    """ASGI-Tools decoding error."""


class ASGINotFound(ASGIError):
    """Raise when http handler not found."""


class ASGIMethodNotAllowed(ASGIError):
    """Raise when http method not found."""


DEFAULT_CHARSET = "utf-8"
BASE_ENCODING = "latin-1"

from http_router import MethodNotAllowed, NotFound  # noqa

from .app import App  # noqa
from .middleware import RequestMiddleware  # noqa
from .middleware import (LifespanMiddleware, ResponseMiddleware, RouterMiddleware,
                         StaticFilesMiddleware)
from .request import Request  # noqa
from .response import ResponseFile  # noqa
from .response import (Response, ResponseError, ResponseHTML, ResponseJSON, ResponseRedirect,
                       ResponseSSE, ResponseStream, ResponseText, ResponseWebSocket, parse_response)
from .view import HTTPView  # noqa

__all__ = (
    "App",
    "HTTPView",
    "LifespanMiddleware",
    "MethodNotAllowed",
    "NotFound",
    "Request",
    "RequestMiddleware",
    "Response",
    "ResponseError",
    "ResponseFile",
    "ResponseHTML",
    "ResponseJSON",
    "ResponseMiddleware",
    "ResponseRedirect",
    "ResponseSSE",
    "ResponseStream",
    "ResponseText",
    "ResponseWebSocket",
    "RouterMiddleware",
    "StaticFilesMiddleware",
    "parse_response",
)

# pylama: ignore=E402



"""ASGI-Tools includes a `asgi_tools.Request` class that gives you a nicer interface onto the
incoming request.
"""

from __future__ import annotations

from http import cookies
from typing import Any, AsyncGenerator, Callable, Dict, Iterator, Optional, Union

from multidict import MultiDict
from yarl import URL

from asgi_tools import DEFAULT_CHARSET, ASGIDecodeError
from asgi_tools._compat import json_loads
from asgi_tools.forms import read_formdata
from asgi_tools.types import TJSON, TASGIReceive, TASGIScope, TASGISend
from asgi_tools.utils import CIMultiDict, parse_headers, parse_options_header


class Request(TASGIScope):
    """Represent a HTTP Request.

    :param scope: HTTP ASGI Scope
    :param receive: an asynchronous callable which lets the application
                    receive event messages from the client

    """

    __slots__ = (
        "scope",
        "receive",
        "send",
        "_is_read",
        "_url",
        "_body",
        "_form",
        "_headers",
        "_media",
        "_cookies",
    )

    def __init__(self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend):
        """Create a request based on the given scope."""
        self.scope = scope
        self.receive = receive
        self.send = send

        self._is_read: bool = False
        self._url: Optional[URL] = None
        self._body: Optional[bytes] = None
        self._form: Optional[MultiDict] = None
        self._headers: Optional[CIMultiDict] = None
        self._media: Optional[Dict[str, str]] = None
        self._cookies: Optional[Dict[str, str]] = None

    def __str__(self) -> str:
        """Return the request's params."""
        scope_type = self.scope["type"]
        if scope_type == "websocket":
            return f"{scope_type} {self.path}"

        return f"{scope_type} {self.method} {self.url.path}"

    def __repr__(self):
        """Represent the request."""
        return f"<Request {self}>"

    def __getitem__(self, key: str) -> Any:
        """Proxy the method to the scope."""
        return self.scope[key]

    def __setitem__(self, key: str, value: Any) -> None:
        """Proxy the method to the scope."""
        self.scope[key] = value

    def __delitem__(self, key: str) -> None:
        """Proxy the method to the scope."""
        del self.scope[key]

    def __iter__(self) -> Iterator[str]:
        """Proxy the method to the scope."""
        return iter(self.scope)

    def __len__(self) -> int:
        """Proxy the method to the scope."""
        return len(self.scope)

    def __getattr__(self, name: str) -> Any:
        """Proxy the request's unknown attributes to scope."""
        return self.scope[name]

    def __copy__(self, **mutations) -> Request:
        """Copy the request to a new one."""
        return Request(dict(self.scope, **mutations), self.receive, self.send)

    @property
    def url(self) -> URL:
        """A lazy property that parses the current URL and returns :class:`yarl.URL` object.

        .. code-block:: python

            request = Request(scope)
            assert str(request.url) == '... the full http URL ..'
            assert request.url.scheme
            assert request.url.host
            assert request.url.query is not None
            assert request.url.query_string is not None

        See :py:mod:`yarl` documentation for further reference.
        """
        if self._url is None:
            scope = self.scope
            host = self.headers.get("host")
            if host is None:
                if "server" in scope:
                    host, port = scope["server"]
                    if port:
                        host = f"{host}:{port}"
                else:
                    host = "localhost"

            self._url = URL.build(
                host=host,
                scheme=scope.get("scheme", "http"),
                encoded=True,
                path=f"{ scope.get('root_path', '') }{ scope['path'] }",
                query_string=scope["query_string"].decode(encoding="ascii"),
            )

        return self._url

    @property
    def headers(self) -> CIMultiDict:
        """A lazy property that parses the current scope's headers, decodes them as strings and
        returns case-insensitive multi-dict :py:class:`multidict.CIMultiDict`.

        .. code-block:: python

            request = Request(scope)

            assert request.headers['content-type']
            assert request.headers['authorization']

        See :py:mod:`multidict` documentation for futher reference.

        """
        if self._headers is None:
            self._headers = parse_headers(self.scope["headers"])
        return self._headers

    @property
    def cookies(self) -> Dict[str, str]:
        """A lazy property that parses the current scope's cookies and returns a dictionary.

        .. code-block:: python

            request = Request(scope)
            ses = request.cookies.get('session')

        """
        if self._cookies is None:
            self._cookies = {}
            cookie = self.headers.get("cookie")
            if cookie:
                for chunk in cookie.split(";"):
                    key, _, val = chunk.partition("=")
                    self._cookies[key.strip()] = cookies._unquote(val.strip())

        return self._cookies

    @property
    def media(self) -> Dict[str, str]:
        """Prepare a media data for the request."""
        if self._media is None:
            conten_type_header = self.headers.get("content-type", "")
            content_type, opts = parse_options_header(conten_type_header)
            self._media = dict(opts, content_type=content_type)

        return self._media

    @property
    def charset(self) -> str:
        """Get an encoding charset for the current scope."""
        return self.media.get("charset", DEFAULT_CHARSET)

    @property
    def query(self) -> MultiDict:
        """A lazy property that parse the current query string and returns it as a
        :py:class:`multidict.MultiDict`.

        """
        return self.url.query

    @property
    def content_type(self) -> str:
        """Get a content type for the current scope."""
        return self.media["content_type"]

    async def stream(self) -> AsyncGenerator:
        """Stream the request's body.

        The method provides byte chunks without storing the entire body to memory.
        Any subsequent calls to :py:meth:`body`, :py:meth:`form`, :py:meth:`json`
        or :py:meth:`data` will raise an error.

        .. warning::
            You can only read stream once. Second call raises an error. Save a readed stream into a
            variable if you need.

        """
        if self._is_read:
            if self._body is None:
                raise RuntimeError("Stream has been read")
            yield self._body

        else:
            self._is_read = True
            message = await self.receive()
            yield message.get("body", b"")
            while message.get("more_body"):
                message = await self.receive()
                yield message.get("body", b"")

    async def body(self) -> bytes:
        """Read and return the request's body as bytes.

        `body = await request.body()`
        """
        if self._body is None:
            self._body = b"".join([chunk async for chunk in self.stream()])

        return self._body

    async def text(self) -> str:
        """Read and return the request's body as a string.

        `text = await request.text()`
        """
        body = await self.body()
        try:
            return body.decode(self.charset or DEFAULT_CHARSET)
        except (LookupError, ValueError) as exc:
            raise ASGIDecodeError("Invalid Encoding") from exc

    async def json(self) -> TJSON:
        """Read and return the request's body as a JSON.

        `json = await request.json()`

        """
        try:
            return json_loads(await self.body())
        except (LookupError, ValueError) as exc:
            raise ASGIDecodeError("Invalid JSON") from exc

    async def form(
        self,
        max_size: int = 0,
        upload_to: Optional[Callable] = None,
        file_memory_limit: int = 1024 * 1024,
    ) -> MultiDict:
        """Read and return the request's multipart formdata as a multidict.

        The method reads the request's stream stright into memory formdata.
        Any subsequent calls to :py:meth:`body`, :py:meth:`json` will raise an error.

        `formdata = await request.form()`

        """
        if self._form is None:
            try:
                self._form = await read_formdata(
                    self, max_size, upload_to, file_memory_limit
                )
            except (LookupError, ValueError) as exc:
                raise ASGIDecodeError("Invalid Encoding") from exc

        return self._form

    async def data(
        self, raise_errors: bool = False
    ) -> Union[str, bytes, MultiDict, TJSON]:
        """The method checks Content-Type Header and parse the request's data automatically.

        `data = await request.data()`

        If `raise_errors` is false (by default) and the given data is invalid (ex. invalid json)
        the request's body would be returned.

        Returns data from :py:meth:`json` for `application/json`, :py:meth:`form` for
        `application/x-www-form-urlencoded`,  `multipart/form-data` and :py:meth:`text` otherwise.
        """
        try:
            if self.content_type in {
                "application/x-www-form-urlencoded",
                "multipart/form-data",
            }:
                return await self.form()

            if self.content_type == "application/json":
                return await self.json()

            return await self.text()

        except ASGIDecodeError:
            if raise_errors:
                raise
            return await self.body()



"""Setup the package."""


import pathlib

# Parse requirements
# ------------------
import pkg_resources


def parse_requirements(path: str) -> "list[str]":
    with pathlib.Path(path).open() as requirements:
        return [str(req) for req in pkg_resources.parse_requirements(requirements)]


# Setup extensions
# ----------------
import os
import sys

from setuptools import Extension

NO_EXTENSIONS = sys.implementation.name != "cpython" or bool(
    os.environ.get("ASGI_TOOLS_NO_EXTENSIONS")
)
EXT_MODULES = (
    []
    if NO_EXTENSIONS
    else [
        Extension(
            "asgi_tools.multipart",
            ["asgi_tools/multipart.c"],
            extra_compile_args=["-O2"],
        ),
        Extension(
            "asgi_tools.forms", ["asgi_tools/forms.c"], extra_compile_args=["-O2"]
        ),
    ]
)


# Setup package
# -------------
from setuptools import setup

setup(
    setup_requires=["wheel"],
    ext_modules=EXT_MODULES,
    install_requires=parse_requirements("requirements/requirements.txt"),
    extras_require={
        "dev": parse_requirements("requirements/requirements-dev.txt"),
        "build": parse_requirements("requirements/requirements-build.txt"),
        "docs": parse_requirements("requirements/requirements-docs.txt"),
        "examples": parse_requirements("requirements/requirements-examples.txt"),
        "orjson": parse_requirements("requirements/requirements-orjson.txt"),
        "ujson": parse_requirements("requirements/requirements-ujson.txt"),
    },
)

# pylama:ignore=E402,D



from typing import ClassVar, Type  # py37

from http_router import Router as HTTPRouter

from . import ASGIError, ASGIMethodNotAllowed, ASGINotFound


class Router(HTTPRouter):
    """Rebind router errors."""

    NotFound: ClassVar[Type[Exception]] = ASGINotFound
    RouterError: ClassVar[Type[Exception]] = ASGIError
    MethodNotAllowed: ClassVar[Type[Exception]] = ASGIMethodNotAllowed



import inspect
from collections.abc import Awaitable
from typing import Optional

from http_router.types import TMethods

from .request import Request
from .router import Router

HTTP_METHODS = {
    "GET",
    "HEAD",
    "POST",
    "PUT",
    "DELETE",
    "CONNECT",
    "OPTIONS",
    "TRACE",
    "PATCH",
}


class HTTPView:
    """Class-based view pattern for handling HTTP method dispatching.

    .. code-block:: python

        @app.route('/custom')
        class CustomEndpoint(HTTPView):

            async def get(self, request):
                return 'Hello from GET'

            async def post(self, request):
                return 'Hello from POST'

        # ...
        async def test_my_endpoint(client):
            response = await client.get('/custom')
            assert await response.text() == 'Hello from GET'

            response = await client.post('/custom')
            assert await response.text() == 'Hello from POST'

            response = await client.put('/custom')
            assert response.status_code == 405

    """

    def __new__(cls, request: Request, **opts):
        """Init the class and call it."""
        self = super().__new__(cls)
        return self(request, **opts)

    @classmethod
    def __route__(
        cls,
        router: Router,
        *paths: str,
        methods: Optional[TMethods] = None,
        **params,
    ):
        """Bind the class view to the given router."""
        view_methods = dict(inspect.getmembers(cls, inspect.isfunction))
        methods = methods or [m for m in HTTP_METHODS if m.lower() in view_methods]
        return router.bind(cls, *paths, methods=methods, **params)

    def __call__(self, request: Request, **opts) -> Awaitable:
        """Dispatch the given request by HTTP method."""
        method = getattr(self, request.method.lower())
        return method(request, **opts)
